from featuretools.computational_backends import calculate_feature_matrix
from featuretools.synthesis import dfs
from sklearn.base import TransformerMixin


class DFSTransformer(TransformerMixin):
    """Transformer using Scikit-Learn interface for Pipeline uses.
    """

    def __init__(self,
                 target_dataframe_name=None,
                 agg_primitives=None,
                 trans_primitives=None,
                 allowed_paths=None,
                 max_depth=2,
                 ignore_dataframes=None,
                 ignore_columns=None,
                 seed_features=None,
                 drop_contains=None,
                 drop_exact=None,
                 where_primitives=None,
                 max_features=-1,
                 verbose=False):
        """Creates Transformer

        Args:

            target_dataframe_name (str): Name of dataframe on which to make
                predictions.

            agg_primitives (list[str or AggregationPrimitive], optional): List
                of Aggregation Feature types to apply.

                    Default: ["sum", "std", "max", "skew", "min", "mean",
                              "count", "percent_true", "num_unique", "mode"]

            trans_primitives (list[str or TransformPrimitive], optional):
                List of Transform Feature functions to apply.

                    Default: ["day", "year", "month", "weekday", "haversine",
                              "num_words", "num_characters"]

            allowed_paths (list[list[str]]): Allowed dataframe paths on which to
                make features.

            max_depth (int) : Maximum allowed depth of features.

            ignore_dataframes (list[str], optional): List of dataframes to
                blacklist when creating features.

            ignore_columns (dict[str -> list[str]], optional): List of
                specific columns within each dataframe to blacklist when
                creating features.

            seed_features (list[:class:`.FeatureBase`]): List of manually
                defined features to use.

            drop_contains (list[str], optional): Drop features
                that contains these strings in name.

            drop_exact (list[str], optional): Drop features that
                exactly match these strings in name.

            where_primitives (list[str or PrimitiveBase], optional):
                List of Primitives names (or types) to apply with where
                clauses.

                    Default:

                        ["count"]

            max_features (int, optional) : Cap the number of generated features
                    to this number. If -1, no limit.

        Example:
            .. ipython:: python

                import featuretools as ft
                import pandas as pd

                from featuretools.wrappers import DFSTransformer
                from sklearn.pipeline import Pipeline
                from sklearn.ensemble import ExtraTreesClassifier

                # Get example data
                train_es = ft.demo.load_mock_customer(return_entityset=True, n_customers=3)
                test_es = ft.demo.load_mock_customer(return_entityset=True, n_customers=2)
                y = [True, False, True]

                # Build pipeline
                pipeline = Pipeline(steps=[
                    ('ft', DFSTransformer(target_dataframe_name="customers",
                                          max_features=2)),
                    ('et', ExtraTreesClassifier(n_estimators=100))
                ])

                # Fit and predict
                pipeline.fit(X=train_es, y=y) # fit on customers in training entityset
                pipeline.predict_proba(test_es) # predict probability of each class on test entityset
                pipeline.predict(test_es) # predict on test entityset

                # Same as above, but using cutoff times
                train_ct = pd.DataFrame()
                train_ct['customer_id'] = [1, 2, 3]
                train_ct['time'] = pd.to_datetime(['2014-1-1 04:00',
                                                   '2014-1-2 17:20',
                                                   '2014-1-4 09:53'])

                pipeline.fit(X=(train_es, train_ct), y=y)

                test_ct = pd.DataFrame()
                test_ct['customer_id'] = [1, 2]
                test_ct['time'] = pd.to_datetime(['2014-1-4 13:48',
                                                  '2014-1-5 15:32'])
                pipeline.predict_proba((test_es, test_ct))
                pipeline.predict((test_es, test_ct))

        """
        self.feature_defs = []
        self.target_dataframe_name = target_dataframe_name
        self.agg_primitives = agg_primitives
        self.trans_primitives = trans_primitives
        self.allowed_paths = allowed_paths
        self.max_depth = max_depth
        self.ignore_dataframes = ignore_dataframes
        self.ignore_columns = ignore_columns
        self.seed_features = seed_features
        self.drop_contains = drop_contains
        self.drop_exact = drop_exact
        self.where_primitives = where_primitives
        self.max_features = max_features
        self.verbose = verbose

    def fit(self, X, y=None):
        """Wrapper for DFS

            Calculates a list of features given a dictionary of dataframes and a list
            of relationships. Alternatively, an EntitySet can be passed instead of
            the dataframes and relationships.

            Args:
                X: (ft.Entityset or tuple): Entityset to calculate features on. If a tuple is
                    passed it can take one of these forms: (entityset, cutoff_time_dataframe),
                    (dataframes, relationships), or ((dataframes, relationships), cutoff_time_dataframe)
                y: (iterable): Training targets

            See Also:
                :func:`synthesis.dfs`
        """
        es, dataframes, relationships, _ = parse_x_input(X)

        self.feature_defs = dfs(entityset=es,
                                dataframes=dataframes,
                                relationships=relationships,
                                target_dataframe_name=self.target_dataframe_name,
                                agg_primitives=self.agg_primitives,
                                trans_primitives=self.trans_primitives,
                                allowed_paths=self.allowed_paths,
                                max_depth=self.max_depth,
                                ignore_dataframes=self.ignore_dataframes,
                                ignore_columns=self.ignore_columns,
                                seed_features=self.seed_features,
                                drop_contains=self.drop_contains,
                                drop_exact=self.drop_exact,
                                where_primitives=self.where_primitives,
                                max_features=self.max_features,
                                features_only=True,
                                verbose=self.verbose)

        return self

    def transform(self, X):
        """Wrapper for calculate_feature_matrix

            Calculates a feature matrix for a the given input data and calculation times.

            Args:
                X: (ft.Entityset or tuple): Entityset to calculate features on. If a tuple is
                    passed it can take one of these forms: (entityset, cutoff_time_dataframe),
                    (dataframes, relationships), or ((dataframes, relationships), cutoff_time_dataframe)

            See Also:
                :func:`computational_backends.calculate_feature_matrix`
        """
        es, dataframes, relationships, cutoff_time = parse_x_input(X)

        X_transformed = calculate_feature_matrix(
            features=self.feature_defs,
            instance_ids=None,
            cutoff_time=cutoff_time,
            entityset=es,
            dataframes=dataframes,
            relationships=relationships,
            verbose=self.verbose)

        return X_transformed

    def get_params(self, deep=True):
        out = {
            'target_dataframe_name': self.target_dataframe_name,
            'agg_primitives': self.agg_primitives,
            'trans_primitives': self.trans_primitives,
            'allowed_paths': self.allowed_paths,
            'max_depth': self.max_depth,
            'ignore_dataframes': self.ignore_dataframes,
            'ignore_columns': self.ignore_columns,
            'seed_features': self.seed_features,
            'drop_contains': self.drop_contains,
            'drop_exact': self.drop_exact,
            'where_primitives': self.where_primitives,
            'max_features': self.max_features,
            'verbose': self.verbose,
        }
        return out


def parse_x_input(X):
    if isinstance(X, tuple):
        if isinstance(X[0], tuple):
            # Input of ((dataframes, relationships), cutoff_time)
            dataframes = X[0][0]
            relationships = X[0][1]
            es = None
            cutoff_time = X[1]
        elif isinstance(X[0], dict):
            # Input of (dataframes, relationships)
            dataframes = X[0]
            relationships = X[1]
            es = None
            cutoff_time = None
        else:
            # Input of (entityset, cutoff_time)
            es = X[0]
            dataframes = None
            relationships = None
            cutoff_time = X[1]
    else:
        # Input of entityset
        es = X
        dataframes = None
        relationships = None
        cutoff_time = None

    return es, dataframes, relationships, cutoff_time



import copy

import numpy as np
import pandas as pd
import pytest
from featuretools.demo.mock_customer import load_mock_customer
from featuretools.wrappers import DFSTransformer
from sklearn.ensemble import ExtraTreesClassifier
from sklearn.impute import SimpleImputer
from sklearn.model_selection import GridSearchCV, cross_val_score
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import FunctionTransformer, StandardScaler


def select_numeric(df):
    return df.select_dtypes(exclude=["object", pd.CategoricalDtype])


@pytest.fixture
def es():
    es = load_mock_customer(
        n_customers=15,
        n_products=15,
        n_sessions=75,
        n_transactions=1000,
        random_seed=0,
        return_entityset=True,
    )
    return es


@pytest.fixture
def es_customer_filtered(es):
    new_es = copy.deepcopy(es)
    customers_df = es["customers"]
    sessions_df = es["sessions"]
    products_df = es["products"]
    transactions_df = es["transactions"]
    customer_ids = [1, 2, 3]
    customers_df = customers_df.loc[customer_ids]
    sessions_df = sessions_df[sessions_df["customer_id"].isin(customer_ids)]
    transactions_df = transactions_df[
        transactions_df["session_id"].isin(sessions_df["session_id"].values)
    ]
    products_df = products_df[
        products_df["product_id"].isin(transactions_df["product_id"].values)
    ]
    new_es.replace_dataframe("customers", customers_df)
    new_es.replace_dataframe("sessions", sessions_df)
    new_es.replace_dataframe("transactions", transactions_df)
    new_es.replace_dataframe("products", products_df)

    return new_es


def get_dataframes_and_relationships(es):
    dataframes = {}
    relationships = []

    for df in es.dataframes:
        dataframes[df.ww.name] = (
            df,
            df.ww.index,
            df.ww.time_index,
            df.ww.logical_types,
        )

    for rel in es.relationships:
        relationships.append(
            (
                rel._parent_dataframe_name,
                rel._parent_column_name,
                rel._child_dataframe_name,
                rel._child_column_name,
            )
        )

    return dataframes, relationships


@pytest.fixture
def df(es):
    df = es["customers"].copy()
    df["target"] = np.random.randint(1, 3, df.shape[0])  # 1 or 2 values
    return df


@pytest.fixture
def pipeline():
    pipeline = Pipeline(
        steps=[
            ("ft", DFSTransformer(target_dataframe_name="customers", max_features=20)),
            ("numeric", FunctionTransformer(select_numeric, validate=False)),
            ("imp", SimpleImputer()),
            ("et", ExtraTreesClassifier(n_estimators=10)),
        ]
    )
    return pipeline


def test_sklearn_transformer_with_entityset(es):
    # Using with transformers
    pipeline = Pipeline(
        steps=[
            ("ft", DFSTransformer(target_dataframe_name="customers")),
            ("numeric", FunctionTransformer(select_numeric, validate=False)),
            ("sc", StandardScaler()),
        ]
    )

    X_train = pipeline.fit(es).transform(es)

    assert X_train.shape[0] == 15


def test_sklearn_transformer_with_dataframes_and_relationships(es):
    # Using with transformers
    pipeline = Pipeline(
        steps=[
            ("ft", DFSTransformer(target_dataframe_name="customers")),
            ("numeric", FunctionTransformer(select_numeric, validate=False)),
            ("sc", StandardScaler()),
        ]
    )
    dataframes, relationships = get_dataframes_and_relationships(es)

    X_train = pipeline.fit((dataframes, relationships)).transform(
        (dataframes, relationships)
    )

    assert X_train.shape[0] == 15


def test_sklearn_estimator_with_entityset(df, es, pipeline):
    # Using with estimator
    pipeline.fit(es, y=df.target.values).predict(es)
    result = pipeline.score(es, df.target.values)

    assert isinstance(result, (float))

    # Pickling / Unpickling Pipeline
    # TODO fix this
    # s = pickle.dumps(pipeline)
    # pipe_pickled = pickle.loads(s)
    # result = pipe_pickled.score(df['customer_id'].values, df.target.values)
    # assert isinstance(result, (float))


def test_sklearn_estimator_with_dataframes_and_relationships(df, es, pipeline):
    # Using with estimator
    dataframes, relationships = get_dataframes_and_relationships(es)
    pipeline.fit((dataframes, relationships), y=df.target.values).predict(
        (dataframes, relationships)
    )
    result = pipeline.score((dataframes, relationships), df.target.values)

    assert isinstance(result, (float))


# cross_val_score cannot split entityset input
@pytest.mark.xfail
def test_sklearn_cross_val_score(df, es, pipeline):
    # Using with cross_val_score
    results = cross_val_score(
        pipeline, X=es, y=df.target.values, cv=2, scoring="accuracy"
    )

    assert isinstance(results[0], (float))
    assert isinstance(results[1], (float))


# GridSearchCV cannot split entityset input
@pytest.mark.xfail
def test_sklearn_gridsearchcv(df, es, pipeline):
    # Using with GridSearchCV
    params = {"et__max_depth": [5, 10]}
    grid = GridSearchCV(estimator=pipeline, param_grid=params, cv=3)
    grid.fit(es, df.target.values)

    assert len(grid.predict(df["customer_id"].values)) == 15


def test_sklearn_cutoff_with_entityset(pipeline, es_customer_filtered):
    # Using cutoff_time to filter data
    ct = pd.DataFrame()
    ct["customer_id"] = [1, 2, 3]
    ct["time"] = pd.to_datetime(["2014-1-1 04:00", "2014-1-1 04:00", "2014-1-1 04:00"])
    ct["label"] = [True, True, False]

    results = pipeline.fit(X=(es_customer_filtered, ct), y=ct.label).predict(
        X=(es_customer_filtered, ct)
    )

    assert len(results) == 3


def test_sklearn_cutoff_with_dataframes_and_relationships(
    pipeline, es_customer_filtered
):
    # Using cutoff_time to filter data
    ct = pd.DataFrame()
    ct["customer_id"] = [1, 2, 3]
    ct["time"] = pd.to_datetime(["2014-1-1 04:00", "2014-1-1 04:00", "2014-1-1 04:00"])
    ct["label"] = [True, True, False]

    dataframes, relationships = get_dataframes_and_relationships(es_customer_filtered)
    results = pipeline.fit(X=((dataframes, relationships), ct), y=ct.label).predict(
        X=((dataframes, relationships), ct)
    )

    assert len(results) == 3


def test_cfm_uses_filtered_target_df_with_entityset(es):
    pipeline = Pipeline(
        steps=[("ft", DFSTransformer(target_dataframe_name="transactions"))]
    )

    train_ids = [1, 2, 3]
    test_ids = [10, 55, 853]

    train_es = filter_transactions(es, ids=train_ids)
    test_es = filter_transactions(es, ids=test_ids)

    fm_train = pipeline.fit_transform(X=train_es)
    assert all(fm_train["sessions.COUNT(transactions)"] == [1, 1, 1])
    assert set(fm_train.index.values) == set(train_ids)

    fm_test = pipeline.transform(test_es)

    assert all(fm_test["sessions.COUNT(transactions)"] == [1, 2, 2])
    assert set(fm_test.index.values) == set(test_ids)


def test_cfm_uses_filtered_target_df_with_dataframes_and_relationships(es):
    pipeline = Pipeline(
        steps=[("ft", DFSTransformer(target_dataframe_name="transactions"))]
    )

    train_ids = [3, 1, 2]
    test_ids = [853, 55, 10]

    train_es = filter_transactions(es, ids=train_ids)
    test_es = filter_transactions(es, ids=test_ids)
    train_dataframes, train_relationships = get_dataframes_and_relationships(train_es)
    test_dataframes, test_relationships = get_dataframes_and_relationships(test_es)

    fm_train = pipeline.fit_transform(X=(train_dataframes, train_relationships))
    assert all(fm_train["sessions.COUNT(transactions)"] == [1, 1, 1])
    assert set(fm_train.index.values) == set(train_ids)

    fm_test = pipeline.transform(X=(test_dataframes, test_relationships))
    assert all(fm_test["sessions.COUNT(transactions)"] == [2, 2, 1])
    assert set(fm_test.index.values) == set(test_ids)


def filter_transactions(es, ids):
    new_es = copy.deepcopy(es)
    customers_df = es["customers"]
    sessions_df = es["sessions"]
    products_df = es["products"]
    transactions_df = es["transactions"]
    transactions_df = transactions_df.loc[ids]
    sessions_df = sessions_df[
        sessions_df["session_id"].isin(transactions_df["session_id"].values)
    ]
    products_df = products_df[
        products_df["product_id"].isin(transactions_df["product_id"].values)
    ]
    customers_df = customers_df[
        customers_df["customer_id"].isin(sessions_df["customer_id"].values)
    ]
    new_es.replace_dataframe("customers", customers_df)
    new_es.replace_dataframe("sessions", sessions_df)
    new_es.replace_dataframe("transactions", transactions_df, already_sorted=True)
    new_es.replace_dataframe("products", products_df)

    return new_es



from os import path

from setuptools import find_packages, setup

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, "README.md")) as f:
    long_description = f.read()

setup(
    name="featuretools_sklearn_transformer",
    version="1.0.0",
    author="Feature Labs, Inc.",
    author_email="support@featurelabs.com",
    license="BSD 3-clause",
    url="http://www.featurelabs.com/",
    python_requires=">=3.7, <4",
    install_requires=open("requirements.txt").readlines(),
    packages=find_packages(),
    description="Featuretools Transformer for Scikit-Learn Pipeline use.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    include_package_data=True,
    entry_points={
        "featuretools_plugin": [
            "wrappers = featuretools_sklearn_transformer",
        ],
    },
)
